/* Taglut : Topological And Geometrical Library - a Useful Toolkit
 * Copyright (C) 2010 Jean-Marie Favreau <jean-marie.favreau@ens-cachan.org>
 *                    CNRS / Univ. Blaise Pascal / IMATI-CNR
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <algorithm>

#include "MeshMap.h"
#include "VectorField.h"
#include "PLPath.h"
#include "Mesh.h"
#include "Exception.h"

using namespace Taglut;


PLPath::PLPath(const PointOnEdge & point_t, bool up, const Mesh & mesh_t, const MeshMap & mMap, const VectorField & vf, double length_t) : mesh(&mesh_t) {
  addPointAtEnd(point_t);
  double currentLength = 0.;
  double limit = up ? mMap.getMaxValue() : mMap.getMinValue();
  while ((length_t < 0) || (currentLength < length_t)) {
    if (mMap.getValue(points.back()) == limit) {
      break;
    }

    PointOnEdge p;
    if (points.back().isVertex()) {
      try {
	p = vf.getNextPoint(mesh_t.point(points.back().getClosestVertex()), up, mesh_t, mMap);
      } catch (Exception e) {
	// end of a path
	break;
      }
    }
    else if(points.size() != 1) {
      const Triangle & t = mesh_t.findOtherTriangle(points.back().getFirstVertex(), points.back().getSecondVertex(),
						  mesh_t.findTriangle(points.back(), points[points.size() - 2]));
      try {
	p = vf.getNextPoint(points.back(), t, up, mesh_t, mMap);
      } catch (Exception e) {
	// end of a path
	break;
      }
    }
    else {
      bool found = true;
      const Triangle & t1 = mesh_t.findTriangle(points.back().getFirstVertex(), points.back().getSecondVertex());
      PointOnEdge p1;
      try {
	p1 = vf.getNextPoint(points.back(), t1, up, mesh_t, mMap);
      } catch (Exception e) {
	found = false;
      }
      double value1 = std::numeric_limits<double>::max();
      if (found) {
	value1 = mMap.getValue(p1);
      }

      const Triangle & t = mesh_t.findOtherTriangle(points.back().getFirstVertex(), points.back().getSecondVertex(), t1);

      found = true;
      try {
	p = vf.getNextPoint(points.back(), t, up, mesh_t, mMap);
      } catch (Exception e) {
	found = false;
      }

      double value = std::numeric_limits<double>::max();
      if (found) {
	value = mMap.getValue(p);
      }

      if ((value == std::numeric_limits<double>::max()) && (value1 == std::numeric_limits<double>::max())) {
	break; // end of a path
      }

      if ((value1 != std::numeric_limits<double>::max()) &&
	  ((value == std::numeric_limits<double>::max()) || (up && (value < value1)) || ((!up) && (value > value1))))
	p = p1;
    }
    if ((up && (mMap.getValue(p) < mMap.getValue(points.back()))) ||
	((!up) && (mMap.getValue(p) > mMap.getValue(points.back())))) {
      break;
    }
    currentLength += p.distance(points.back());
    addPointAtEnd(p);
  }
}

PLPath & PLPath::addPointAtBeginning(const PointOnEdge & point) {
  assert(mesh != NULL);
  if (points.size() != 0) {
#ifndef NDEBUG
    try {
      if ((point.getLocation() != 0.) && (point.getLocation() != 1.) &&
	  (points.front().getLocation() != 0.) && (points.front().getLocation() != 1.)) {
	(*mesh).findTriangle(point, points.front());
      }
    }
    catch (Exception e) {
      throw Exception("addPointAtEnd(1): the new point is not connected to the path");
    }
#endif
    points.push_front(point);
  }
  else {
    points.push_front(point);
  }
  return *this;
}

PLPath & PLPath::addPointAtEnd(const PointOnEdge & point) {
  assert(mesh != NULL);
  if (points.size() != 0) {
#ifndef NDEBUG
    try {
      if ((point.getLocation() != 0.) && (point.getLocation() != 1.) &&
	  (points.back().getLocation() != 0.) && (points.back().getLocation() != 1.)) {
	(*mesh).findTriangle(point, points.back());
      }
    }
    catch (Exception e) {
      throw Exception("addPointAtEnd(1): the new point is not connected to the path");
    }
#endif
    points.push_back(point);
  }
  else {
    points.push_back(point);
  }
  return *this;
}

PLPath & PLPath::append(const PLPath & path) {
  assert(mesh != NULL);
#ifndef NDEBUG
  if ((points.size() != 0) && (path.size() != 0)) {
    try {
      if ((path.front().getLocation() != 0.) && (path.front().getLocation() != 1.) &&
	  (points.back().getLocation() != 0.) && (points.back().getLocation() != 1.)) {
	(*mesh).findTriangle(path.front(), points.back());
      }
    }
    catch (Exception e) {
      throw Exception("addPointAtEnd(1): the new point is not connected to the path");
    }
  }
#endif

  points.insert(points.end(), path.begin(), path.end());

  return *this;
}



double PLPath::length() const {
  if (points.size() <= 1)
    return 0.0;

  double result = 0.;
  std::deque<PointOnEdge>::const_iterator ppred = points.begin();
  for(std::deque<PointOnEdge>::const_iterator p = ++(points.begin()); p != points.end(); ++p, ++ppred)
    result += (*ppred).distance(*p);

  return result;
}

PLPath & PLPath::reversePath() {
  std::reverse(points.begin(), points.end());
  return *this;
}

PLPath PLPath::computeIntegralLine(const PointOnEdge & point, bool up,
				     const Mesh & mesh,
				     const MeshMap & mMap, const VectorField & vf,
				     double length) {
  return PLPath(point, up, mesh, mMap, vf, length);
}

std::vector<PLPath> PLPath::getIntegralLinesFromSaddlePoint(const Point3D & point,
							    const Mesh & mesh, const MeshMap & mMap, const VectorField & vf,
							    double length, bool twoSides) {
  std::vector<PLPath> result;
  assert(mMap.isSaddlePoint(point.getId()));

  double valuePoint = mMap[point.getId()];
  VertexID pUp[2] = {mesh.getNbPoints(), mesh.getNbPoints()};
  VertexID pDown[2] = {mesh.getNbPoints(), mesh.getNbPoints()};
  unsigned char idSide = 0;
  unsigned char idInP = 0;

  // first computes the next points on integral lines in both sides
  double diff = 0;
  for(std::deque<VertexID>::const_iterator nb = point.getNeighbours().begin();
      nb != point.getNeighbours().end(); ++nb) {
    double v = mMap[*nb];
    if ((nb != point.getNeighbours().begin()) && (diff * (v - valuePoint) < 0)) {
      ++idSide;
      if (idSide == 5)
	throw Exception("getIntegralLinesFromSaddlePoint(5): this method is not able to manage with non regular saddle points.");
      if (idSide == 2)
	idInP = 1;
      if (idSide == 4)
	idInP = 0;
    }
    diff = v - valuePoint;
    if (diff > 0) {
      if ((pUp[idInP] == mesh.getNbPoints()) || (v > mMap[pUp[idInP]])) {
	pUp[idInP] = *nb;
      }
    }
    else if (diff < 0) {
      if ((pDown[idInP] == mesh.getNbPoints()) || (v < mMap[pDown[idInP]])) {
	pDown[idInP] = *nb;
      }
    }
    else {
      std::cout << "getIntegralLinesFromSaddlePoint(5): flat edge detected" << std::endl;
    }
  }
  if ((pUp[0] == mesh.getNbPoints()) || (pUp[1] == mesh.getNbPoints()) ||
      (pDown[0] == mesh.getNbPoints()) || (pDown[1] == mesh.getNbPoints()))
    throw Exception("getIntegralLinesFromSaddlePoint(5): a point is missing in the neighbourhood of the saddle point.");

  if (twoSides) {
    result.push_back(PLPath(PointOnEdge(mesh.point(pUp[0]), point), true, mesh, mMap, vf, length - point.distance(mesh.point(pUp[0]))));
    result.back().addPointAtBeginning(point);
    result.push_back(PLPath(PointOnEdge(mesh.point(pUp[1]), point), true, mesh, mMap, vf, length - point.distance(mesh.point(pUp[1]))));
    result.back().addPointAtBeginning(point);
    result.push_back(PLPath(PointOnEdge(mesh.point(pDown[0]), point), false, mesh, mMap, vf, length - point.distance(mesh.point(pDown[0]))));
    result.back().addPointAtBeginning(point);
    result.push_back(PLPath(PointOnEdge(mesh.point(pDown[1]), point), false, mesh, mMap, vf, length - point.distance(mesh.point(pDown[1]))));
    result.back().addPointAtBeginning(point);
  }
  else {
    // then compute the level sets from each found points
    bool up = true;
    if (mMap[pUp[1]] > mMap[pUp[0]]) {
      PLPath ls = mMap.computeLevelSetCC(mesh.point(pUp[0]));
      if (!ls.intersects(point.getId(), pUp[1]))
        up = false;
    }
    else {
      PLPath ls = mMap.computeLevelSetCC(mesh.point(pUp[1]));
      if (!ls.intersects(point.getId(), pUp[0]))
        up = false;
    }

    if (!up) {
      if (mMap[pDown[1]] > mMap[pDown[0]]) {
        PLPath ls = mMap.computeLevelSetCC(mesh.point(pDown[1]));
        if (!ls.intersects(point.getId(), pDown[0]))
  	throw Exception("getIntegralLinesFromSaddlePoint(5): cannot find the valid direction.");
      }
      else {
        PLPath ls = mMap.computeLevelSetCC(mesh.point(pDown[0]));
        if (!ls.intersects(point.getId(), pDown[1]))
	  throw Exception("getIntegralLinesFromSaddlePoint(5): cannot find the valid direction.");
      }
    }

    // then compute the integral lines
    const VertexID * pp = up ? pUp : pDown;
    double d1 = length - point.distance(mesh.point(pp[0]));
    if (d1 < 0.) d1 = 0.;
    double d2 = length - point.distance(mesh.point(pp[1]));
    if (d2 < 0.) d2 = 0.;
    if (length < 0.)
      d1 = d2 = -1.;
    result.push_back(PLPath(PointOnEdge(mesh.point(pp[0]), point), up, mesh, mMap, vf, d1));
    result.back().addPointAtBeginning(point);
    result.push_back(PLPath(PointOnEdge(mesh.point(pp[1]), point), up, mesh, mMap, vf, d2));
    result.back().addPointAtBeginning(point);
  }

  return result;

}

bool PLPath::intersects(VertexID v1, VertexID v2) const {

  for(std::deque<PointOnEdge>::const_iterator p = points.begin(); p != points.end(); ++p)
    if ((*p).isVertex()) {
      const VertexID c = (*p).getClosestVertex();
      if ((v1 == c) || (v2 == c))
	return true;
    }
    else {
      if (((v1 == (*p).getFirstVertex()) && (v2 == (*p).getSecondVertex())) ||
	  ((v2 == (*p).getFirstVertex()) && (v1 == (*p).getSecondVertex())))
	return true;
    }
  return false;
}

bool PLPath::intersects(const PLPath & path) const {
  PLPath::const_iterator predp = path.begin();
  for(PLPath::const_iterator p = path.begin() + 1; p != path.end(); ++p, ++predp)
    if (intersects(*predp, *p)) {
      return true;
    }
  return false;
}



bool PLPath::intersects(const PointOnEdge & v1, const PointOnEdge & v2) const {
  if (points.size() == 1) {
    const PointOnEdge p = points.front();
    if (p.isVertex()) {
      if ((v1 == p) || (v2 == p))
	return true;
      else
	return false;
    } else {
      if ((v1 == p) || (v2 == p))
	return true;
      else if (p.inSameEdge(v1) && p.inSameEdge(v2)) {
	const double pP = p.getLocation();
	const double pV1 = v1.getLocationOnEdge(p);
	const double pV2 = v2.getLocationOnEdge(p);
	return ((pV1 >= pP) && (pP >= pV2)) || ((pV1 <= pP) && (pP <= pV2));
      }
      else
	return false;
    }
  }
  else {
    std::deque<PointOnEdge>::const_iterator predp = points.begin();
    assert(mesh != NULL);
    for(std::deque<PointOnEdge>::const_iterator p = points.begin() + 1; p != points.end(); ++p, ++predp) {
      if ((*mesh).intersects(*predp, *p, v1, v2)) {
	return true;
      }
    }
    return false;
  }

}

std::vector<Coord3D> PLPath::getCoordsCutPath(double value, bool direction, const MeshMap & mMap) const {
  std::vector<Coord3D> result;
  if (points.size() == 0)
    return result;

  // check for the valid starting point
  if ((direction && (value > mMap.getValue(points.front()))) ||
      (!direction && (value < mMap.getValue(points.front())))) {
    // starting from the beginning
    std::deque<PointOnEdge>::const_iterator predp = points.begin();
    std::deque<PointOnEdge>::const_iterator p = points.begin();
    for(; ((p != points.end()) &&
	   ((direction && (value > mMap.getValue(*p))) ||
	    (!direction && (value < mMap.getValue(*p))))); predp = p, ++p)
      result.push_back(*p);
    if (p != points.end()) {
      const PointOnEdge & p1 = *predp;
      const PointOnEdge & p2 = *p;
      assert(!(p1 == p2));
      assert(((mMap.getValue(p2) >= value) && (value >= mMap.getValue(p1))) ||
	     ((mMap.getValue(p1) >= value) && (value >= mMap.getValue(p2))));
      double v = (mMap.getValue(p2) - value) / (mMap.getValue(p2) - mMap.getValue(p1));
      assert((v >= 0) && (v <= 1));
      result.push_back(p2 + (p1 - p2) * v);
    }
  }
  else if ((direction && (value > mMap.getValue(points.back()))) ||
	   (!direction && (value < mMap.getValue(points.back())))) {
    // starting from the end
    std::deque<PointOnEdge>::const_reverse_iterator predp = points.rbegin();
    std::deque<PointOnEdge>::const_reverse_iterator p = points.rbegin();
    for(; ((p != points.rend()) &&
	   ((direction && (value > mMap.getValue(*p))) ||
	    (!direction && (value < mMap.getValue(*p))))); predp = p, ++p)
      result.push_back(*p);
    if (p != points.rend()) {
      const PointOnEdge & p1 = *predp;
      const PointOnEdge & p2 = *p;
      assert(!(p1 == p2));
      assert(((mMap.getValue(p2) >= value) && (value >= mMap.getValue(p1))) ||
	     ((mMap.getValue(p1) >= value) && (value >= mMap.getValue(p2))));
      double v = (mMap.getValue(p2) - value) / (mMap.getValue(p2) - mMap.getValue(p1));
      assert((v >= 0) && (v <= 1));
      result.push_back(p2 + (p1 - p2) * v);
    }

    std::reverse(result.begin(), result.end());
  }
  else {
    throw Exception("getCoordsCutPath(3): both begin and end sides of the path are outside of the wanted area");
  }

  return result;
}

PLPath & PLPath::cutPath(double value, bool direction, const MeshMap & mMap) {
  if (points.size() == 0)
    return *this;

  std::deque<PointOnEdge> newPath;
  // check for the valid starting point
  if ((direction && (value > mMap.getValue(points.front()))) ||
      (!direction && (value < mMap.getValue(points.front())))) {
    // starting from the beginning
    for(std::deque<PointOnEdge>::const_iterator p = points.begin(); ((p != points.end()) &&
								     ((newPath.size() == 0) ||
								      (direction && (value > mMap.getValue(newPath.back()))) ||
								      (!direction && (value < mMap.getValue(newPath.back()))))); ++p)
      newPath.push_back(*p);
  }
  else if ((direction && (value > mMap.getValue(points.back()))) ||
	   (!direction && (value < mMap.getValue(points.back())))) {
    // starting from the end
    for(std::deque<PointOnEdge>::const_reverse_iterator p = points.rbegin(); ((p != points.rend()) &&
									      ((newPath.size() == 0) ||
									       (direction && (value > mMap.getValue(newPath.back()))) ||
									       (!direction && (value < mMap.getValue(newPath.back()))))); ++p)
      newPath.push_front(*p);
  }
  else {
    throw Exception("cutPath(3): both begin and end sides of the path are outside of the wanted area");
  }
  points = newPath;

  return *this;
}

double PLPath::getIntersectionValue(const PointOnEdge & v1, const PointOnEdge & v2) const {
  if (size() < 2)
    throw Exception("getIntersectionCoord(1): there is no intersection");
  assert(mesh != NULL);
  PLPath::const_iterator previous = begin();
  Line3D v1v2(v1, v2 - v1);
  for(PLPath::const_iterator p = begin() + 1; p != end(); previous = p, ++p) {
    if ((*mesh).intersects(*previous, *p, v1, v2)) {
      Line3D pp(*previous, (*p) - (*previous));
      return v1v2.getIntersectionLocation(pp);
    }
  }

  throw Exception("getIntersectionValue(2): there is no intersection");
}

Coord3D PLPath::getIntersectionCoord(const PLPath & path) const{
  if (size() < 2)
    throw Exception("getIntersectionCoord(1): there is no intersection");
  PLPath::const_iterator previous = path.begin();
  for(PLPath::const_iterator p = path.begin() + 1; p != path.end(); previous = p, ++p)
    if (intersects(*previous, *p))
      return getIntersectionCoord(*previous, *p);
  throw Exception("getIntersectionCoord(1): there is no intersection");
}


double PLPath::getIntersectionValue(const PLPath & path) const {
  double result = 0.;
  if (size() < 2)
    throw Exception("getIntersectionValue(1): there is no intersection (current path too small)");
  if (path.size() < 2)
    throw Exception("getIntersectionValue(1): there is no intersection (path too small)");
  assert(mesh != NULL);

  PLPath::const_iterator previous = begin();
  for(PLPath::const_iterator p = begin() + 1; p != end(); previous = p, ++p)
    if (path.intersects(*previous, *p)) {
      double d = path.getIntersectionValue(*previous, *p);
      assert(d >= -1e-7);
      assert(d <= 1. + 1e-7);
      if (d <= 0.) d = 0.;
      if (d >= 1.) d = 1.;
      return result + d * (*previous).distance(*p);
    }
    else
      result += (*previous).distance(*p);

  throw Exception("getIntersectionValue(1): no intersection found");
}

std::pair<double, double> PLPath::getExtremaValuesBetweenPoints(double p1, double p2) const {
  assert(isLoop());
  const double l = length();

  if (p1 == p2)
    throw Exception("getExtremaValuesBetweenPoints(2): same location");
  if ((p1 > l) || (p1 < 0.))
    throw Exception("getExtremaValuesBetweenPoints(2): first value out of range");
  if ((p2 > l) || (p2 < 0.))
    throw Exception("getExtremaValuesBetweenPoints(2): second value out of range");

  // order the points
  double pp1, pp2;
  if (p1 < p2) {
    pp1 = p1;
    pp2 = p2;
  }
  else {
    pp1 = p2;
    pp2 = p1;
  }

  if ((pp1 == 0.) && (pp2 == l))
    throw Exception("getExtremaValuesBetweenPoints(2): same location (loop).");

  // compute the extrema
  const double r1 = (pp2 + pp1) / 2;

  double r2 = (pp2 + l + pp1) / 2;
  while(r2 > l)
    r2 -= l;

  return std::pair<double, double>(r1, r2);
}

Coord3D PLPath::getCoordFromValue(double input) const {
  if (size() == 0)
    throw Exception("getCoordFromValue(1): empty path");
  if ((input == 0.) || (size() == 1))
    return front();

  double current = 0.;
  PLPath::const_iterator previous = begin();
  for(PLPath::const_iterator p = begin() + 1; p != end(); previous = p, ++p) {
    const double next = current + (*previous).distance(*p);
    if (next < input)
      current = next;
    else {
      const double r = (input - current) / (next - current);
      return (*previous) + ((*p) - (*previous)) * r;
    }
  }

  throw Exception("getCoordFromValue(1): geodesic coordinate out of range");
}

std::string PLPath::toString() const {
  std::ostringstream s;
  if (size() == 0)
    return s.str();

  s << front().toString();
  for(PLPath::const_iterator p = begin() + 1; p != end(); ++p)
    s << "->" << (*p).toString();

  if (isLoop())
    s << "->";

  return s.str();
}

double PLPath::distanceMod(double p1, double p2, double length) {
  if (p1 > p2) {
    const double d1 = p1 - p2;
    const double d2 = p2 + length - p1;
    if (d1 < d2)
      return d1;
    else
      return d2;
  }
  else {
    const double d1 = p2 - p1;
    const double d2 = p1 + length - p2;
    if (d1 < d2)
      return d1;
    else
      return d2;
  }

}

double PLPath::middleMod(double p1, double p2, double length) {
  if (p1 > p2)
    return middleMod(p2, p1, length);
  if (fabs(p1 - p2) < fabs(p1 + length - p2))
    return (p1 + p2) / 2;
  else {
    const double m = (p1 + p2 + length) / 2;
    return m > length ? m - length : m;
  }
}

double PLPath::meanMod(const std::vector<double> & values, double length, double outside) {
  if (values.size() == 0)
    throw Exception("meanMod(2): empty values");
  if (values.size() == 1)
    return values.front();

  double sum = 0.;
  for(std::vector<double>::const_iterator p = values.begin(); p != values.end(); ++p) {
      sum += *p;
      if (*p < outside)
	sum += length;
  }
  sum /= values.size();
  return sum > length ? sum - length : sum;
}


const PointOnEdge & PLPath::getClosestPointOnEdge(double value) const {
  if (size() == 0)
    throw Exception("getClosestPointOnEdge(1): empty path");
  if ((value == 0.) || (size() == 1))
    return front();

  double v = 0.;
  PLPath::const_iterator pred = begin();
  for(PLPath::const_iterator p = begin() + 1; p != end(); pred = p, ++p) {
    v += (*pred).distance(*p);
    if (value <= v) {
      if (value < v - (*pred).distance(*p) / 2)
	return *pred;
      else
	return *p;
    }
  }
  throw Exception("getClosestPointOnEdge(1): value out of range");
}

std::pair<double, PointOnEdge> PLPath::getMiddleByAngle(std::vector<Coord3D> & points_l, bool local) const {
  if (size() == 0)
    throw Exception("getMiddleByAngle(1): empty path");
  if (points_l.size() < 2)
    throw Exception("getMiddleByAngle(1): it needs more than one point.");

  // optimal ratio
  const double target = 1. / points_l.size();

  double dist = -1;
  PLPath::const_iterator result;
  for(PLPath::const_iterator pOnPath = begin(); pOnPath != end(); ++pOnPath) {
    double min = std::numeric_limits<double>::max();
    double max = -1.;
    double sum = 0.;
    std::vector<Coord3D>::const_iterator pred = points_l.begin() + (points_l.size() - 1);
    for(std::vector<Coord3D>::const_iterator p = points_l.begin(); p != points_l.end(); pred = p, ++p) {
      const double a = (*pOnPath).angle(*pred, *p);
      assert(a > 0.);
      sum += a;
      if (a < min) min = a;
      if (a > max) max = a;
    }
    assert(min >= 0.);
    assert(max > 0.);
    min /= sum;
    max /= sum;

    double v = max - target;
    if (target - min > v) v = target - min;

    if ((dist == -1) || (v < dist)) {
      dist = v;
      result = pOnPath;
    }
    else
      if (local)
	break;
  }
  assert(dist != -1);
  return std::pair<double, PointOnEdge>(dist, *result);
}

PLPath PLPath::operator+ (const PLPath & path) const {
  assert(back() == path.front());
  PLPath result(*this);
  if (path.size() >= 2)
    for(std::deque<PointOnEdge>::const_iterator p = path.points.begin() + 1; p != path.points.end(); ++p)
      result.addPointAtEnd(*p);
  return result;
}

void PLPath::adjustForAdjacency() {
  std::deque<PointOnEdge>::iterator ppred = points.begin();
  for(std::deque<PointOnEdge>::iterator p = points.begin(); p != points.end(); ++p) {
    if (ppred != p) {
      if ((*p).isVertex() && (*ppred).isVertex()) {
	if ((*p).getFirstVertex() == (*p).getSecondVertex()) {
	  (*p).setSecondVertex((*ppred).getClosestVertex());
	  (*p).setLocation(0.);
	}
      }
    }
    ppred = p;
  }
}


std::vector<Coord3D> PLPath::subdivise(double s) const {
  if (s == 0.)
    throw Exception("subdivise(1): zero value.");
  std::vector<Coord3D> result;
  if (size() == 0)
    return result;

  result.push_back(front());
  PLPath::const_iterator pred = begin();
  for(PLPath::const_iterator p = begin() + 1; p != end(); ++p, ++pred) {
    const double d = (*pred).distance(*p);
    unsigned int nbsteps = floor(d / s);
    if (nbsteps != 0) {
      Coord3D pp(*pred);
      Coord3D v((*p - *pred).normalize() * d / nbsteps);
      for(unsigned int i = 0; i != nbsteps; ++i) {
	pp += v;
	result.push_back(pp);
      }
    }
    result.push_back(*p);
    assert(result.size() >= 2);
    assert(result.back().distance(result[result.size() - 2]) <= s);
  }
  return result;
}
