/* Taglut : Topological And Geometrical Library - a Useful Toolkit
 * Copyright (C) 2011 Jean-Marie Favreau <J-Marie.Favreau@u-clermont1.fr>
 *                    Université d'Auvergne Clermont1
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include<algorithm>
#include "MTiling.h"
#include "BarycentricMapping.h"

using namespace Taglut;


MTile::MTile(MeshPart & mPart, unsigned int i) : id(i),
						 v(mPart.getInsidePoint()),
						 mesh(&(mPart.getMesh())),
						 genus(mPart.getGenus()),
						 nbBoundaries(mPart.getNbBoundaries()) {
}

MTile::MTile(const MTile & mt) : multipoints(mt.multipoints), id(mt.id),
				 neighbours(mt.neighbours), v(mt.v),
				 mesh(mt.mesh), genus(mt.genus), nbBoundaries(mt.nbBoundaries) {
}

MTile & MTile::operator=(MTile & mt) {
  multipoints = mt.multipoints;
  id = mt.id;
  neighbours = mt.neighbours;
  v = mt.v;
  mesh = mt.mesh;
  genus = mt.genus;
  nbBoundaries = mt.nbBoundaries;
  return *this;
}

void MTile::addMultiPoint(VertexID multipoint) {
  if(find(multipoints.begin(), multipoints.end(), multipoint) == multipoints.end())
    multipoints.push_back(multipoint);
}

void MTile::addNeighbour(const MTile & mtile) {
  neighbours.push_back(mtile.getId());
}


std::vector<VertexID> MTile::getMultiPointsSameLocation(const Coord3D & coord, const Mesh & mesh) {
  std::vector<VertexID> result;
  for(std::vector<VertexID>::const_iterator m = multipoints.begin(); m != multipoints.end(); ++m)
    if (coord == mesh.point(*m))
      result.push_back(*m);

  return result;
}

void MTile::removeTwiceConnections() {
  std::vector<unsigned int> result;
  std::map<unsigned int, unsigned int> nbC;
  for(std::vector<unsigned int>::const_iterator n = neighbours.begin(); n != neighbours.end(); ++n)
    nbC[*n] += 1;

  for(std::map<unsigned int, unsigned int>::iterator nn = nbC.begin(); nn != nbC.end(); ++nn) {
    assert(!(((*nn).second) & 1)); // even number of similar connections
    result.insert(result.end(), (*nn).second / 2, (*nn).first);
  }

  neighbours = result;
}

MTiling::MTiling(Mesh & m) : mesh(m) {
  std::deque<MeshPart> cc = m.getConnectedComponents();
  unsigned int i = 0;
  // build the connected components
  for(std::deque<MeshPart>::iterator c = cc.begin(); c != cc.end(); ++c, ++i)
    mtiles.push_back(MTile(*c, i));

  // set the flags according to the CC
  mesh.setPointFlag(-1);
  MeshManipulator mManip(mesh);
  for(std::vector<MTile>::const_iterator t = mtiles.begin(); t != mtiles.end(); ++t) {
    assert(mesh.point((*t).getInsidePoint()).getFlag() == -1);
    mManip.computeCCUsingPointsSimple((*t).getInsidePoint(), -1, (*t).getId());
  }

  // compute multipoints
  std::map<Coord3D, std::vector<VertexID>, Coord3D::ltCoord> mPoints = mesh.computeMultiPointsByCoords();

  // add multipoints and neighbours to the tiles
  for(std::map<Coord3D, std::vector<VertexID>, Coord3D::ltCoord>::const_iterator mpl = mPoints.begin(); mpl != mPoints.end(); ++mpl) {
    for(std::vector<VertexID>::const_iterator mp = (*mpl).second.begin(); mp != (*mpl).second.end(); ++mp) {
      assert(mtiles[mesh.point(*mp).getFlag()].getId() == (unsigned int)(mesh.point(*mp).getFlag()));
      mtiles[mesh.point(*mp).getFlag()].addMultiPoint(*mp);
    }
    setNeighbourConnections((*mpl).second);
  }

  // remove twin connections (each connection has been added twice because of the structure)
  for(std::vector<MTile>::iterator t = mtiles.begin(); t != mtiles.end(); ++t)
    (*t).removeTwiceConnections();
}

void MTiling::setNeighbourConnections(const std::vector<VertexID> & mPoints) {
  // add connections using multipoints
  for(std::vector<VertexID>::const_iterator p = mPoints.begin(); p != mPoints.end(); ++p) {
    unsigned int pFlag = mesh.point(*p).getFlag();
    VertexID nbB[2];
    unsigned int n = 0;
    // first get the two neighbours
    for(std::deque<VertexID>::const_iterator nbP = mesh.point(*p).getNeighbours().begin();
	nbP != mesh.point(*p).getNeighbours().end(); ++nbP)
      if (mesh.isBoundaryEdge(*nbP, *p)) {
	if (n == 2)
	  throw Exception("Bad boundary structure");
	nbB[n] = *nbP;
	++n;
      }

    // then check if the other multipoints have same location
    for(std::vector<VertexID>::const_iterator pp = p + 1; pp != mPoints.end(); ++pp) {
      for(std::deque<VertexID>::const_iterator nbPP = mesh.point(*pp).getNeighbours().begin();
	  nbPP != mesh.point(*pp).getNeighbours().end(); ++nbPP) {
	const Point3D & nPPC = mesh.point(*nbPP);
	if(nPPC.getIsBoundary() && (mesh.point(nbB[0]).sameLocation(nPPC) ||
				    mesh.point(nbB[1]).sameLocation(nPPC))) {
	  addConnection(mesh.point(*pp).getFlag(), pFlag);
	  break;
	}
      }
    }
  }
}


std::vector<unsigned int> MTiling::getNeighbours(unsigned int idTile,
						 unsigned int g,
						 unsigned int nbb,
						 unsigned int nbMP) const {
  std::vector<unsigned int> result;
  const std::vector<unsigned int> & neighbours = mtiles[idTile].getNeighbours();

  for(std::vector<unsigned int>::const_iterator n = neighbours.begin(); n != neighbours.end(); ++n) {
    const MTile & mt = mtiles[*n];
    if ((mt.getGenus() == g) && (mt.getNbBoundaries() == nbb) && (mt.getNbMultiPoints() == nbMP))
      result.push_back(*n);
  }
  return result;
}

void MTiling::adjustQuads(MTile & mt1, MTile & mt2) {
  assert(mt1.hasNeighbour(mt2));
  assert((mt1.getGenus() == 0) && (mt1.getNbBoundaries() == 1) && (mt1.getNbMultiPoints() == 2));
  assert((mt2.getGenus() == 0) && (mt2.getNbBoundaries() == 1) && (mt2.getNbMultiPoints() == 2));

  std::pair<std::vector<VertexID>, std::vector<VertexID> > bd = getCommonBorder(mt1, mt2);
  assert(bd.first.size() != 0);
  assert(bd.first.size() == bd.second.size());
  double l = mesh.getLength(bd.first);
  assert(l != 0.);
  assert(l == mesh.getLength(bd.second));

  double l1 = l * 2. / 7.;
  double l2 = l * 5. / 7.;

  std::pair<VertexID, VertexID> pts1 = mesh.getClosestVertices(bd.first, l1, l2);
  std::pair<VertexID, VertexID> pts2 = mesh.getClosestVertices(bd.second, l1, l2);
  mt1.addMultiPoint(pts1.first);
  mt1.addMultiPoint(pts1.second);
  mt2.addMultiPoint(pts2.first);
  mt2.addMultiPoint(pts2.second);
  assert(mt1.getNbMultiPoints() == 4);
  assert(mt2.getNbMultiPoints() == 4);
}


std::pair<std::vector<VertexID>, std::vector<VertexID> > MTiling::getCommonBorder(MTile & mt1, MTile & mt2) const {
  std::pair<std::vector<VertexID>, std::vector<VertexID> > result;

  if (!mt1.hasNeighbour(mt2))
    return result;

  for(std::vector<VertexID>::const_iterator mmt1 = mt1.getMultiPoints().begin();
      mmt1 != mt1.getMultiPoints().end(); ++mmt1) {
    std::vector<VertexID> mpts2 = mt2.getMultiPointsSameLocation(mesh.point(*mmt1), mesh);
    if (mpts2.size() != 0) {
      for(std::vector<VertexID>::const_iterator mmt2 = mpts2.begin(); mmt2 != mpts2.end(); ++mmt2) {
	VertexID nextmmt1 = mesh.getNextBPoint(*mmt1);
	VertexID nextmmt2 = mesh.getNextBPoint(*mmt2);
	if (!mesh.point(nextmmt1).sameLocation(mesh.point(nextmmt2))) {
	  nextmmt2 = mesh.getNextBPoint(*mmt2, nextmmt2);
	  if (!mesh.point(nextmmt1).sameLocation(mesh.point(nextmmt2))) {
	    nextmmt1 = mesh.getNextBPoint(*mmt1, nextmmt1);
	    if (!mesh.point(nextmmt1).sameLocation(mesh.point(nextmmt2))) {
	      nextmmt2 = mesh.getNextBPoint(*mmt2, nextmmt2);
	      if (!mesh.point(nextmmt1).sameLocation(mesh.point(nextmmt2))) {
		continue;
	      }
	    }
	  }
	}
	assert(mesh.point(nextmmt1).sameLocation(mesh.point(nextmmt2)));
	result.first.push_back(*mmt1);
	result.second.push_back(*mmt2);
	result.first.push_back(nextmmt1);
	result.second.push_back(nextmmt2);
	while(!mt1.hasMultiPoint(result.first.back())) {
	  result.first.push_back(mesh.getNextBPoint(result.first.back(), result.first[result.first.size() - 2]));
	  result.second.push_back(mesh.getNextBPoint(result.second.back(), result.second[result.second.size() - 2]));
	  assert(result.first.size() == result.second.size());
	  assert(mesh.point(result.first.back()).sameLocation(mesh.point(result.second.back())));
	}
	return result;
      }
    }
  }

  return result;
}

void MTiling::adjustQuads() {
  std::vector<bool> checked(mtiles.size(), false);


  for(std::vector<MTile>::iterator t = mtiles.begin(); t != mtiles.end(); ++t)
    if ((!checked[(*t).getId()]) && ((*t).getGenus() == 0) && ((*t).getNbBoundaries() == 1) && ((*t).getNbMultiPoints() == 2)) {
      checked[(*t).getId()] = true;
      std::vector<unsigned int> neighbours = getNeighbours((*t).getId(), 0, 1, 2);
      if (neighbours.size() != 1)
	throw Exception("adjustQuads(): configuration not handled. Abort.");
      if (getNeighbours(neighbours.front(), 0, 1, 2).size() != 1)
	throw Exception("adjustQuads(): configuration not handled. Abort.");
      adjustQuads(*t, mtiles[neighbours.front()]);
      checked[neighbours.front()] = true;
    }
}

Mapping2D3D MTiling::parameterize() {
  Mapping2D3D mapping(mesh);

  for(std::vector<MTile>::const_iterator t = mtiles.begin(); t != mtiles.end(); ++t) {
    MeshPart mPart(mesh, (*t).getInsidePoint());
    Mesh ccMesh = mPart.buildCropMesh();
    MeshPart::IDTranslator idTranslator(mPart);

    if (((*t).getGenus() != 0) || (*t).getNbBoundaries() != 1)
      throw Exception("Wrong topology configuration");

    Mapping2D3D rMapping(ccMesh);
    if ((*t).getNbMultiPoints() == 4) {
      QuadParameterizationFloater mParam;
      assert(mPart.hasPoint((*t).getMultiPoint(0)));
      assert(mPart.hasPoint((*t).getMultiPoint(1)));
      assert(mPart.hasPoint((*t).getMultiPoint(2)));
      assert(mPart.hasPoint((*t).getMultiPoint(3)));
      rMapping = mParam.parameterize(ccMesh,
				     idTranslator.g2l((*t).getMultiPoint(0)),
				     idTranslator.g2l((*t).getMultiPoint(1)),
				     idTranslator.g2l((*t).getMultiPoint(2)),
				     idTranslator.g2l((*t).getMultiPoint(3)));
    }
    else {
      DiscParameterizationFloater mParam;
      rMapping = mParam.parameterize(ccMesh);
    }

    VertexID lID = 0;
    for(Mapping2D3D::const_iterator p2D = rMapping.begin(); p2D != rMapping.end(); ++p2D, ++lID)
      mapping[idTranslator.l2g(lID)] = *p2D;
  }

  // the mesh is flatten
  mapping.setFlatten();

  return mapping;
}

