/* Taglut : Topological And Geometrical Library - a Useful Toolkit
 * Copyright (C) 2011 Jean-Marie Favreau <J-Marie.Favreau@u-clermont1.fr>
 *                    Université d'Auvergne Clermont1
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef M_TILING
#define M_TILING

#include<vector>

#include "Mesh.h"
#include "MeshPart.h"
#include "Mapping2D3D.h"
#include "IDTypes.h"

namespace Taglut {

  /**
     An MTile is a structure to describe a connected component in a mesh resulting from
     a cutting. Specific properties as multipoints are contained in this structure.
   */
  class MTile {
  private:
    /** list of multipoint on the boundary, i.e. points
	which location is the same as points in other MTiles (or in another
	part of the current tile) */
    std::vector<VertexID> multipoints;
    /** id of the current MTile */
    unsigned int id;
    /** the neighbours of the current tile */
    std::vector<unsigned int> neighbours;
    /** a vertex of the mesh contained in this tile */
    VertexID v;
    /** the manipulated mesh */
    Mesh * mesh;
    /** genus of the tile */
    unsigned int genus;
    /** number of boundaries of the tile */
    unsigned int nbBoundaries;

  public:
    /** default constructor
	@param m The MeshPart we are focussing on
	@param i ID of the current MTile in the MTiling */
    MTile(MeshPart & m, unsigned int i);

    /** copy constructor */
    MTile(const MTile & mt);

#ifndef SWIG
    /** copy operator */
    MTile & operator=(MTile & mt);
#endif

    /** add a multipoint on the boundary */
    void addMultiPoint(VertexID multipoint);

    /** add a neighbour tile */
    void addNeighbour(const MTile & mtile);

    /** accessor */
    inline VertexID getInsidePoint() const { return v; }

    /** accessor */
    inline unsigned int getId() const { return id; }

    /** accessor */
    inline unsigned int getNbMultiPoints() const { return multipoints.size(); }

    /** accessor */
    inline VertexID getMultiPoint(unsigned int id) const {
      assert(id < multipoints.size());
      return multipoints[id];
    }

    /** accessor */
    inline const std::vector<unsigned int> & getNeighbours() const { return neighbours; }

    /** return true if the current tile have the given one as neighbour */
    inline bool hasNeighbour(unsigned int idnb) const {
      for(std::vector<unsigned int>::const_iterator nb = neighbours.begin();
	  nb != neighbours.end(); ++nb)
	if (idnb == *nb)
	  return true;
      return false;
    }

    /** return true if the current tile have the given one as neighbour */
    inline bool hasNeighbour(const MTile & mt) const {
      return hasNeighbour(mt.getId());
    }

    /** return true if the current tile have the given multipoint */
    inline bool hasMultiPoint(VertexID p) const {
      for(std::vector<VertexID>::const_iterator mp = multipoints.begin();
	  mp != multipoints.end(); ++mp)
	if (p == *mp)
	  return true;
      return false;
    }

    /** accessor */
    inline unsigned int getGenus() const { return genus; }

    /** accessor */
    inline unsigned int getNbBoundaries() const { return nbBoundaries; }

    /** accessor */
    inline const std::vector<VertexID> & getMultiPoints() const { return multipoints; }

    /** return the multipoints of the current MTile located at the given location \p coord */
    std::vector<VertexID> getMultiPointsSameLocation(const Coord3D & coord, const Mesh & mesh);

    /** assuming that each connection has been added 2 time, it removes the twins */
    void removeTwiceConnections();
  };


  /** this structure is able to manipulate meshes separated by connected components, produced by cuttings:
      some vertices are in the same location... */
  class MTiling {
  private:
    /** description of the tiles */
    std::vector<MTile> mtiles;
    /** the manipulated mesh */
    Mesh & mesh;

    void setNeighbourConnections(const std::vector<VertexID> & mPoints);

    inline void addConnection(unsigned int id1, unsigned int id2) {
      assert(id1 < mtiles.size());
      assert(id2 < mtiles.size());
      mtiles[id1].addNeighbour(mtiles[id2]);
      mtiles[id2].addNeighbour(mtiles[id1]);
    }

    /** return the number of neighbour tiles of the given one with the
	required topology.
	@param idTile ID of the given tile
	@param g genus of the wanted neighbours
	@param nbb number of boundaries of the wanted neighbours
	@param nbMP number of wanted multipoints on the neighbour
	@return ID list of the corresponding tiles
    */
    std::vector<unsigned int> getNeighbours(unsigned int idTile,
					    unsigned int g = 0,
					    unsigned int nbb = 1,
					    unsigned int nbMP = 2) const;


    /** assuming that the two tiles are adjacent discs with only two multipoints,
	this function adds two multipoints points on the shared boundary to create quadrangles */
    void adjustQuads(MTile & mt1, MTile & mt2);
  public:
    /** constructor using a mesh */
    MTiling(Mesh & m);

    MTiling(const MTiling & mtiling) : mtiles(mtiling.mtiles), mesh(mtiling.mesh) { }

    /** assuming that all the tiles are 0-genus, and all the tiles are quad, except adjacent pairs of 2-side tiles,
	it add multipoints in the common boundary of each pair to obtain quadrangles */
    void adjustQuads();

    /** compute a 2D mapping of each tile, using the following rule:
	if the tile is a quadrangle, the unfolding is produced in a rectangle, using a Floater
	parameterization. Otherwise, a Floater parameterization with a circle boundary is applied, */
    Mapping2D3D parameterize();


    /** given two tiles, it return two paths that corresponds to a common border */
    std::pair<std::vector<VertexID>, std::vector<VertexID> > getCommonBorder(MTile & mt1, MTile & mt2) const;
  };

}

#endif
