/* Taglut : Topological And Geometrical Library - a Useful Toolkit
 * Copyright (C) 2010 Jean-Marie Favreau <jean-marie.favreau@ens-cachan.org>
 *                    CNRS / Univ. Blaise Pascal / IMATI-CNR
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */


#include "Mesh.h"
#include "MeshPLCut.h"
#include <algorithm>
#include <set>
#include <vector>
#include "CylinderParameterization.h"
#include "GeodesicDistance.h"

using namespace Taglut;


void MeshPLCut::sortBySide(Mesh & cMesh, MeshPart & cc, const std::vector<VertexID> & mpInsideCC, VertexID sideA[2], VertexID sideB[2]) {
  // first select a boundary of the cylinder
  std::vector<VertexID> boundary = cc.getBoundaryFromBPoint(mpInsideCC.front());


  // then build the two set of points
  cMesh.setPointFlag(0);
  for(std::vector<VertexID>::const_iterator b = boundary.begin(); b != boundary.end(); ++b)
    cMesh.point(*b).setFlag(1);
  unsigned int nbA, nbB;
  nbA = nbB = 0;
  for(std::vector<VertexID>::const_iterator mp = mpInsideCC.begin(); mp != mpInsideCC.end(); ++mp) {
    if (cMesh.point(*mp).getFlag() == 1) {
      if (nbA == 2)
	throw Exception("buildRectangularCutFromCylindricalTilingSimple(): wrong number of multipoints on A face.");
      sideA[nbA] = *mp;
      ++nbA;
    }
    else {
      if (nbB == 2)
	throw Exception("buildRectangularCutFromCylindricalTilingSimple(): wrong number of multipoints on B face.");
      sideB[nbB] = *mp;
      ++nbB;
    }
  }
  cMesh.setPointFlag(0);

}




Mesh MeshPLCut::cutMesh(const Mesh & mesh, const std::vector<PLPath> & paths) {
  Mesh result;

  // clear the data structures
  trianglesNotClean.clear();
  newPoints.clear();

  // add the original points
  result.addPoints(mesh.getPoints(), true);

  // compute the state of each triangle (is it a "clean" triangle, or a triangle that have to be split)
  // and add the new points
  const TriangleID nbTriangles = mesh.getNbTriangles();
  bool triangleClean[mesh.getNbTriangles()];
  bool * tc = triangleClean;
  for(TriangleID i = 0; i != nbTriangles; ++i, ++tc)
    (*tc) = true;

  for(std::vector<PLPath>::const_iterator path = paths.begin(); path != paths.end(); ++path) {
    PLPath corrected(*path);
    corrected.adjustForAdjacency();
    PLPath::const_iterator pOepred = corrected.begin();
    for(PLPath::const_iterator pOe = corrected.begin(); pOe != corrected.end(); pOepred = pOe, ++pOe) {
      // add the new edges
      if (pOepred != pOe) {
	addNewPoint(result, *pOepred);
	addNewPoint(result, *pOe);
	assert(newPoints[*pOepred].getNbCopies() != 0);
	assert(newPoints[*pOe].getNbCopies() != 0);

	if ((*pOe).inSameEdge(*pOepred)) {
	  std::pair<VertexID, VertexID> points = (*pOe).getEdge(*pOepred);
	  try {
	    const Triangle & t1 = mesh.findTriangle(points.first, points.second);
	    addNonCleanTriangle(t1, *pOepred, *pOe);
	    triangleClean[t1.getId()] = false;
	    const Triangle & t2 = mesh.findOtherTriangle(points.first, points.second, t1);
	    addNonCleanTriangle(t2, *pOepred, *pOe);
	    triangleClean[t2.getId()] = false;
	  }
	  catch (Exception e) {
	  }
	}
	else {
	  const Triangle & t = mesh.findTriangle(*pOe, *pOepred);
	  addNonCleanTriangle(t, *pOepred, *pOe);
	  triangleClean[t.getId()] = false;
	}


	// if the current point is a splitted vertex, all the 1-star has to be managed
	if ((*pOe).isVertex() && (newPoints[*pOe].getNbCopies() == 2) && (mesh.point((*pOe).getClosestVertex()).getIsBoundary())) {
	  const Point3D & p = mesh.point((*pOe).getClosestVertex());
	  for(std::deque<TriangleID>::const_iterator t = p.getTriangles().begin(); t != p.getTriangles().end(); ++t) {
	    triangleClean[*t] = false;
	    trianglesNotClean[*t] = mesh.triangle(*t);
	  }
	}

	// if the pred point is a splitted vertex, all the 1-star has to be managed
	if ((*pOepred).isVertex() && (newPoints[*pOepred].getNbCopies() == 2)) {
	  const Point3D & p = mesh.point((*pOepred).getClosestVertex());
	  for(std::deque<TriangleID>::const_iterator t = p.getTriangles().begin(); t != p.getTriangles().end(); ++t) {
	    triangleClean[*t] = false;
	    trianglesNotClean[*t] = mesh.triangle(*t);
	  }
	}
      }
    }
  }

  // add the "clean" triangles
  tc = triangleClean;
  for(Mesh::const_triangle_iterator tt = mesh.getTriangles().begin(); tt != mesh.getTriangles().end(); ++tt, ++tc)
    if (*tc)
      result.addTriangleFull(*tt);


  // compute the new ids for old vertices in each triangle
  for(std::map<PointOnEdge, CreatedPoint>::iterator p = newPoints.begin(); p != newPoints.end(); ++p)
    if ((*p).first.isVertex()) {
      const Point3D & pp = mesh.point((*p).first.getClosestVertex());
      std::vector<VertexID>::const_iterator pNewIds = (*p).second.getNewIds().begin();

      if ((*p).second.getNewIds().size() > 1) {
	// only neighbours are ordered in the 1-star. To get the triangles well ordered, we have to walk on the neighbours
	std::deque<VertexID>::const_iterator nbLast = pp.getNeighbours().begin() + (pp.getNbNeighbours() - 1);
	for(std::deque<VertexID>::const_iterator nb = pp.getNeighbours().begin(); nb != pp.getNeighbours().end(); ++nb) {
	  if ((pp.getNbNeighbours() == 2) && pp.getIsBoundary() && (nbLast == pp.getNeighbours().begin()))
	    // specific configuration: on the boundary, and a single triangle. Do not use it two time
	    break;
	  try {
	    const Triangle & t = mesh.findTriangle(pp.getId(), *nbLast, *nb);
	    if (trianglesNotClean.find(t.getId()) != trianglesNotClean.end()) {
	      // count the number of new edges on this triangle, except the one located on (pp.getId(), *nb)
	      unsigned int nbPtsInNbEdge = trianglesNotClean[t.getId()].getNbCutInEdgeThroughCorner(pp.getId(), *nb);
	      assert(nbPtsInNbEdge <= 1);
	      unsigned int nbPtsInTriangle = trianglesNotClean[t.getId()].getNbCutThroughCorner(pp.getId());
	      unsigned int nbPtsInNbLastEdge = trianglesNotClean[t.getId()].getNbCutInEdgeThroughCorner(pp.getId(), *nbLast);
	      assert(nbPtsInNbLastEdge <= 1);
	      assert(nbPtsInTriangle >= nbPtsInNbLastEdge + nbPtsInNbEdge);
	      if (nbPtsInTriangle != 0) {
		assert((*p).second.getNewIds().size() >= nbPtsInTriangle);
		pNewIds = pNewIds + nbPtsInNbLastEdge;
		if (pNewIds == (*p).second.getNewIds().end()) {
		  pNewIds = (*p).second.getNewIds().begin();
		}

		std::vector<VertexID>::const_iterator pNewIdsEnd = pNewIds + (nbPtsInTriangle - nbPtsInNbEdge - nbPtsInNbLastEdge);
		if (pNewIdsEnd == (*p).second.getNewIds().end()) {
		  pNewIdsEnd = (*p).second.getNewIds().begin();
		}
		trianglesNotClean[t.getId()].setNewIdsOnCorner(pp.getId(), pNewIds, pNewIdsEnd,
							       (*p).second.getNewIds().end(), (*p).second.getNewIds().begin());
		pNewIds = pNewIdsEnd;
	      }
	      else {
		trianglesNotClean[t.getId()].setNewIdsOnCorner(pp.getId(), pNewIds, pNewIds,
							       (*p).second.getNewIds().end(), (*p).second.getNewIds().begin());
	      }
	    }
	    else {
	      assert(pp.getId() == *pNewIds);
	    }
	  }
	  catch (Exception e) {
	    // a jump: pp is a point on the boundary...
	    pNewIds++;
	    if (pNewIds == (*p).second.getNewIds().end())
	      pNewIds = (*p).second.getNewIds().begin();
	  }
	  nbLast = nb;
	}
      }
    }

  // FIXME(0): solve this limitation
#ifndef NDEBUG
  for(std::map<PointOnEdge, CreatedPoint>::const_iterator p = newPoints.begin(); p != newPoints.end(); ++p)
    if ((!(*p).first.isVertex()) && (*p).second.getNewIds().size() != 2)
      Exception("Multi cut on an edge: not able to manage with this configuration");
#endif


  // for each of the "non-clean" triangles, add the corresponding point to the triangle
  for(std::map<TriangleID, TriangleWithCut>::iterator t = trianglesNotClean.begin(); t != trianglesNotClean.end(); ++t)
    addNewTriangles(result, (*t).second, mesh);

  assert(result.point(0).getNbNeighbours() != 0);
  result.reorderNeighbours();
  result.initIsBoundaryFlag();

  return result;
}

void MeshPLCut::addNonCleanTriangle(const Triangle & t, const PointOnEdge & pOe1, const PointOnEdge & pOe2) {
  std::map<TriangleID, TriangleWithCut>::iterator tit = trianglesNotClean.find(t.getId());
  if (tit == trianglesNotClean.end())
    trianglesNotClean[t.getId()] = t;

  assert(trianglesNotClean[t.getId()] == t);
  assert((pOe1.isVertex() && (t.hasPoint(pOe1.getFirstVertex()) || t.hasPoint(pOe1.getSecondVertex()))) ||
	 (!pOe1.isVertex() && t.hasPoint(pOe1.getFirstVertex()) && t.hasPoint(pOe1.getSecondVertex())));
  assert((pOe2.isVertex() && (t.hasPoint(pOe2.getFirstVertex()) || t.hasPoint(pOe2.getSecondVertex()))) ||
	 (!pOe2.isVertex() && t.hasPoint(pOe2.getFirstVertex()) && t.hasPoint(pOe2.getSecondVertex())));
  trianglesNotClean[t.getId()].addCut(pOe1, pOe2);
}

void MeshPLCut::addNewTriangles(Mesh & mesh, TriangleWithCut & triangle, const Mesh & oMesh) {
  if (triangle.nbCuts() == 0) {
    mesh.addTriangleFull(Triangle(triangle.getIdOnCorner(0),
				  triangle.getIdOnCorner(1),
				  triangle.getIdOnCorner(2)));
  }
  else {
    const std::vector<std::pair<PointOnEdge, PointOnEdge> > cuts = triangle.getCuts();
    // create a new multitriangle with a first child (corresponding to an alias in the new data structure)
    MultiTriangle splittedTriangles(triangle, triangle.popFirstIDOnCorner(0), triangle.popFirstIDOnCorner(1), triangle.popFirstIDOnCorner(2));


    for(std::vector<std::pair<PointOnEdge, PointOnEdge> >::const_iterator cut = cuts.begin(); cut != cuts.end(); ++cut) {
      if (!((*cut).first.isVertex() && (*cut).second.isVertex())) {
	const std::pair<EdgeSplitter, EdgeSplitter> cutNewIds = getSplitters(*cut, triangle, oMesh, splittedTriangles);
	splittedTriangles.addCut(*cut, cutNewIds, triangle, mesh);
      }
    }
    splittedTriangles.addInMesh(mesh);
  }
}

void MeshPLCut::addNewPoint(Mesh & mesh, const PointOnEdge & pOe) {

  // count the number of copies
  newPoints[pOe].addNewCopy();

  if ((pOe.isVertex()) && (newPoints[pOe].getNbCopies() == 1)) {
    newPoints[pOe].addNewID(pOe.getClosestVertex());
    if (mesh.point(pOe.getClosestVertex()).getIsBoundary()) {
      newPoints[pOe].addNewCopy();
      newPoints[pOe].addNewID(mesh.addPoint(Point3D(pOe.getX(), pOe.getY(), pOe.getZ())));
    }
  }
  else {
    newPoints[pOe].addNewID(mesh.addPoint(Point3D(pOe.getX(), pOe.getY(), pOe.getZ())));
  }
}


bool MeshPLCut::MultiTriangle::containedInChildren(const PointOnEdge & point, const EdgeSplitter & es, bool first) const {
  assert(isSplit());

  if (point.isVertex()) {
    assert(!(*splitter).isVertex());
    if ((*eSplitter).hasVertex(es.getV1()) && (*eSplitter).hasVertex(es.getV2()))
      return true; // in the opposite corner
    else if ((*eSplitter).hasVertex(es.getV1())) {
      // VertexID vPoint = (*eSplitter).getOtherVertex(es.getV1());
      const VertexID v1 = (*eSplitter).getV1();
      const VertexID v2 = (*eSplitter).getV2();
      return v1 > v2 || !first;
    }
    else {
      assert((*eSplitter).hasVertex(es.getV2()));
      // VertexID vPoint = (*eSplitter).getOtherVertex(es.getV2());
      const VertexID v1 = (*eSplitter).getV1();
      const VertexID v2 = (*eSplitter).getV2();
      return v2 > v1 || !first;
    }
  }
  else {
    if ((*eSplitter) == es) // the given point is the splitter
      return true;
    else if (point.inSameEdge(*splitter)) {
      if (point == *splitter) // the current point is the splitter (with other ids)
	return true;
      bool after = point.getLocationOnEdge(*splitter);
      if (after) { // es is in the (*splitter).getSecondVertex() side, i.e. in the (*eSplitter).getV2() side
	return (*eSplitter).getV2() > (*eSplitter).getV1() || !first;
      }
      else { // es is in the (*splitter).getFirstVertex() side, i.e. in the (*eSplitter).getV1() side
	return (*eSplitter).getV1() > (*eSplitter).getV2() || !first;
      }
    }
    else {
      const VertexID v1 = (*eSplitter).getV1();
      const VertexID v2 = (*eSplitter).getV2();
      assert(hasPoint(v1));
      assert(hasPoint(v2));

      // not in the same edge (in the original mesh) but with v1 and v2 as extremities
      // means that its the opposite corner
      if (es.hasVertex(v1) && es.hasVertex(v2))
	return true;
      assert(es.hasVertex(getOtherPoint(v1, v2)));
      if (es.hasVertex(v1)) {
	return (first && (v1 > v2)) || (!first && (v1 < v2));
      }
      else {
	assert(es.hasVertex(v2));
	return (first && (v2 > v1)) || (!first && (v2 < v1));
      }
    }
  }
}


void MeshPLCut::MultiTriangle::addPointOnEdge(const PointOnEdge & point, VertexID v1, VertexID v2, VertexID middle) {
  assert(isFinal());
  assert(hasPoint(v1));
  assert(hasPoint(v2));
  // create the children
  MultiTriangle * mm1 = new MultiTriangle(*this);
  MultiTriangle * mm2 = new MultiTriangle(*this);
  // add it to the structure
  mt1 = mm1;
  mt2 = mm2;
  assert((*mt1).isFinal());
  assert((*mt2).isFinal());
  // update the current triangle
  splitter = new PointOnEdge(point);
  eSplitter = new EdgeSplitter(v1, v2, middle, middle);
  // update the children
  if (v1 > v2) {
    (*mt1).replacePoint(v2, middle);
    (*mt2).replacePoint(v1, middle);
  }
  else {
    (*mt1).replacePoint(v1, middle);
    (*mt2).replacePoint(v2, middle);
  }

}

void MeshPLCut::MultiTriangle::addPointOnEdgeFullSplit(const PointOnEdge & point, const EdgeSplitter & ids, const EdgeSplitter & idCorner) {
  assert(ids.getId1() != ids.getId2());
  assert(idCorner.getId1() != idCorner.getId2());
  assert(!point.isVertex());
  assert(((ids.getV1() == idCorner.getV1()) && ((ids.getV2() == idCorner.getV2()))) ||
	 ((ids.getV2() == idCorner.getV1()) && ((ids.getV1() == idCorner.getV2()))));
  assert(isFinal());
  assert(hasPoint(ids.getV1()));
  assert(hasPoint(ids.getV2()));
  // create the children
  MultiTriangle * mm1 = new MultiTriangle(*this);
  MultiTriangle * mm2 = new MultiTriangle(*this);
  // add it to the structure
  mt1 = mm1;
  mt2 = mm2;
  assert((*mt1).isFinal());
  assert((*mt2).isFinal());
  // update the current triangle
  splitter = new PointOnEdge(point);
  eSplitter = new EdgeSplitter(ids);
  VertexID oCorner = getOtherPoint(idCorner.getV1(), idCorner.getV2());

  // update the children
  if (ids.getV1() > ids.getV2()) {
    (*mt1).replacePoint(ids.getV1(), ids.getId2());
    (*mt2).replacePoint(ids.getV2(), ids.getId1());
  }
  else {
    (*mt1).replacePoint(ids.getV2(), ids.getId1());
    (*mt2).replacePoint(ids.getV1(), ids.getId2());
  }

  if ((*mt1).hasPoint(idCorner.getV1())) {
    assert((*mt2).hasPoint(idCorner.getV2()));
    assert((*mt2).hasPoint(oCorner));
    (*mt1).replacePoint(oCorner, idCorner.getId1());
    (*mt2).replacePoint(oCorner, idCorner.getId2());
  }
  else {
    assert((*mt1).hasPoint(idCorner.getV2()));
    assert((*mt2).hasPoint(idCorner.getV1()));
    assert((*mt2).hasPoint(oCorner));
    (*mt1).replacePoint(oCorner, idCorner.getId2());
    (*mt2).replacePoint(oCorner, idCorner.getId1());
  }

}

void MeshPLCut::MultiTriangle::addCut(const std::pair<PointOnEdge, PointOnEdge> & cut,
				      const std::pair<EdgeSplitter, EdgeSplitter> & ids,
				      const Triangle & t,
				      const Mesh & mesh) {

  if (isAlias()) {
    // only replace points of the extremities
    (*mt1).addCut(cut, std::pair<EdgeSplitter, EdgeSplitter>(ids.first.getReplacePoints(*this, *mt1, false),
							     ids.second.getReplacePoints(*this, *mt1, false)),
		  t, mesh);
  }
  else if (isSplit()) {

    // if it's an edge cut and the middle point has been already added, break
    if (cut.first.inSameEdge(cut.second) && ((cut.first == *splitter) || (cut.second == *splitter))) {
      return;
    }

    // if the current splitter is one of the two points, we only apply a full split with the other one
    if ((cut.first == *splitter) && (isFinalSplit())) {
      addPointOnEdgeFullSplitChildren(cut, ids);
      return;
    }
    if ((cut.second == *splitter) && (isFinalSplit())) {
      addPointOnEdgeFullSplitChildren(std::pair<PointOnEdge, PointOnEdge>(cut.second, cut.first),
				      std::pair<EdgeSplitter, EdgeSplitter>(ids.second, ids.first));
      return;
    }

    bool added = false;
    if (containedInFirst(cut.first, ids.first) &&
	containedInFirst(cut.second, ids.second)) {
      (*mt1).addCut(cut, std::pair<EdgeSplitter, EdgeSplitter>(ids.first.getReplacePoints(*this, *mt1), ids.second.getReplacePoints(*this, *mt1)),
		    t, mesh);
      added = true;
    }
    if (containedInSecond(cut.first, ids.first) &&
	containedInSecond(cut.second, ids.second)) {
      (*mt2).addCut(cut, std::pair<EdgeSplitter, EdgeSplitter>(ids.first.getReplacePoints(*this, *mt2), ids.second.getReplacePoints(*this, *mt2)),
		    t, mesh);
      added = true;
    }
    if (!added)
      throw Exception("Crossing of cuts in the inner part of a triangle. Situation not handled.");
  }
  else {
    assert(isFinal());
    if (cut.first.isVertex()) {
      if (cut.second.isVertex()) {
	// get the original vertices
	const VertexID v1 = cut.first.getClosestVertex();
	const VertexID v2 = cut.second.getClosestVertex();
	assert(v1 != v2);
	const VertexID v3 = t.getOtherPoint(v1, v2);
	// get the triangle location according to this new cut
	const bool after1 = mesh.point(v1).neighboursValidOrder(v2, v3);
	const bool after2 = mesh.point(v2).neighboursValidOrder(v1, v3);
	VertexID newv1, newv2;
	if (after1)
	  newv1 = ids.first.getIdFromOtherSide(v2); // the one in the v3 side
	else
	  newv1 = ids.first.getIdFromSide(v2);
	if (after2)
	  newv2 = ids.second.getIdFromOtherSide(v1); // the one in the v3 side
	else
	  newv2 = ids.second.getIdFromSide(v1);
      }
      else {
	if (cut.first.inSameEdge(cut.second)) {
	  VertexID newID;
	  if (cut.first.getClosestVertex() == cut.second.getFirstVertex())
	    newID = ids.second.getId2();
	  else {
	    assert(cut.first.getClosestVertex() == cut.second.getSecondVertex());
	    newID = ids.second.getId1();
	  }
	  addPointOnEdge(cut.second, ids.second.getV1(), ids.second.getV2(), newID);
	}
	else
	  addPointOnEdgeFullSplit(cut.second, ids.second, ids.first);
      }
    }
    else {
      if (cut.second.isVertex()) {
	if (cut.first.inSameEdge(cut.second)) {
	  VertexID newID;
	  if (cut.second.getClosestVertex() == cut.first.getFirstVertex())
	    newID = ids.first.getId2();
	  else {
	    assert(cut.second.getClosestVertex() == cut.first.getSecondVertex());
	    newID = ids.first.getId1();
	  }
	  addPointOnEdge(cut.first, ids.first.getV1(), ids.first.getV2(), newID);
	}
	else
	  addPointOnEdgeFullSplit(cut.first, ids.first, ids.second);
      }
      else {
	addPointOnEdge(cut.first, ids.first.getV1(), ids.first.getV2(), ids.first.getId1());
	addPointOnEdgeFullSplitChildren(cut, ids);
      }
    }
  }
}

void MeshPLCut::MultiTriangle::addPointOnEdgeFullSplitChildren(const std::pair<PointOnEdge, PointOnEdge> & cut,
							       const std::pair<EdgeSplitter, EdgeSplitter> & ids) {
  assert((splitter != NULL) && ((*splitter) == cut.first));
  if (containedInFirst(cut.second, ids.second)) {
    EdgeSplitter newIDEdge(ids.second);
    newIDEdge.replacePoints(*this, *mt1);
    EdgeSplitter newIDCorner(ids.first);
    newIDCorner.adjustSamePoints(newIDEdge);
    (*mt1).addPointOnEdgeFullSplit(cut.second, newIDEdge, newIDCorner);
    // adjust (*mt2) if required
    for(unsigned int i = 0; i < 3; ++i)
      if (newIDCorner.hasVertex((*mt2).getVertexId(i))) {
	(*mt2).replacePoint(ids.first.getId1(), newIDCorner.getIdFromSide((*mt2).getVertexId(i)));
	break;
      }
  }
  else {
    assert(containedInSecond(cut.second, ids.second));
    EdgeSplitter newIDEdge(ids.second);
    newIDEdge.replacePoints(*this, *mt2);
    EdgeSplitter newIDCorner(ids.first);
    newIDCorner.adjustSamePoints(newIDEdge);
    (*mt2).addPointOnEdgeFullSplit(cut.second, newIDEdge, newIDCorner);
    // adjust (*mt1) if required
    for(unsigned int i = 0; i < 3; ++i)
      if (newIDCorner.hasVertex((*mt1).getVertexId(i))) {
	(*mt1).replacePoint(ids.first.getId1(), newIDCorner.getIdFromSide((*mt1).getVertexId(i)));
	break;
      }
  }
}

void MeshPLCut::MultiTriangle::addInMesh(Mesh & mesh) {
  if (isSplit()) {
    (*mt1).addInMesh(mesh);
    (*mt2).addInMesh(mesh);
  }
  else if (!isFinal())
    (*mt1).addInMesh(mesh);
  else {
    mesh.addTriangleFull(*this);
  }
}

std::pair<MeshPLCut::EdgeSplitter, MeshPLCut::EdgeSplitter> MeshPLCut::getSplitters(const std::pair<PointOnEdge, PointOnEdge> & cut,
										    const Triangle & t, const Mesh & mesh,
										    const MeshPLCut::MultiTriangle & mTriangle) {
  return std::pair<EdgeSplitter, EdgeSplitter>(getSplitter(cut.first, t, mesh, mTriangle), getSplitter(cut.second, t, mesh, mTriangle));
}

MeshPLCut::EdgeSplitter MeshPLCut::getSplitter(const PointOnEdge & cut, const Triangle & t, const Mesh & mesh,
					       const MeshPLCut::MultiTriangle & mTriangle) {
  assert(mesh.getNbPoints() > cut.getFirstVertex());
  assert(mesh.getNbPoints() > cut.getSecondVertex());
  assert(t.hasPoint(cut.getFirstVertex()) || t.hasPoint(cut.getSecondVertex()));
  if (cut.isVertex()) {
    VertexID v0 = cut.getClosestVertex();
    VertexID v1 = t.getOtherPoint(v0);
    VertexID v2 = t.getOtherPoint(v0, v1);
    // we have to reorder v1, v2 according to the ordering arround v0
    if (!mesh.point(v0).neighboursValidOrder(v1, v2)) {
      assert(mesh.point(v0).neighboursValidOrder(v2, v1));
      VertexID vtemp = v1;
      v1 = v2;
      v2 = vtemp;
    }

    TriangleWithCut & tCut = trianglesNotClean[t.getId()];
    VertexID v0bis = tCut.popIDOnCorner(t.getVertexPositionById(v0));
    assert(mTriangle.isAlias());
    VertexID v0in = mTriangle.getAlias(v0);
    //assert(v0bis != v0in);
    return EdgeSplitter(v1, v2, v0in, v0bis);
  }
  else {
    VertexID v1 = cut.getFirstVertex();
    VertexID v2 = cut.getSecondVertex();
    assert(v1 != v2);

    const CreatedPoint & cp = newPoints[cut];
    assert(cp.getNewIds().size() == 2); // FIXME(0): a limitation
    VertexID vv1 = cp.getNewIds().front();
    VertexID vv2 = cp.getNewIds().back();
    assert(vv1 != vv2);
    return EdgeSplitter(v1, v2, vv1, vv2).reorder();
  }
}


MeshPLCut::EdgeSplitter & MeshPLCut::EdgeSplitter::replacePoints(const VTriangle & t1, const VTriangle & t2, bool full) {
  v1 = t1.hasPoint(v1) ? t2.getVertexId(t1.getVertexPositionById(v1)) : v1;
  v2 = t1.hasPoint(v2) ? t2.getVertexId(t1.getVertexPositionById(v2)) : v2;
  if (full) {
    id1 = t1.hasPoint(id1) ? t2.getVertexId(t1.getVertexPositionById(id1)) : id1;
    id2 = t1.hasPoint(id2) ? t2.getVertexId(t1.getVertexPositionById(id2)) : id2;
  }

  return *this;
}

MeshPLCut::EdgeSplitter & MeshPLCut::EdgeSplitter::replacePoints(VertexID iid1, VertexID iid2, bool full) {
  v1 = (v1 == iid1) ? iid2 : v1;
  v2 = (v2 == iid1) ? iid2 : v2;

  if (full) {
    id1 = (id1 == iid1) ? iid2 : id1;
    id2 = (id2 == iid1) ? iid2 : id2;
  }

  return *this;
}

MeshPLCut::EdgeSplitter & MeshPLCut::EdgeSplitter::adjustSamePoints(const EdgeSplitter & es) {
  if (v1 == es.getV1()) {
    v2 = es.getV2();
  }
  else if (v1 == es.getV2()) {
    v2 = es.getV1();
  }
  else if (v2 == es.getV1()) {
    v1 = es.getV2();
  }
  else {
    assert(v2 == es.getV2());
    v1 = es.getV1();
  }

  return *this;
}

Mesh MeshPLCut::cutMeshInQuadranglesFromCylindricalTiling(Mesh & mesh) {
  std::vector<PLPath> paths;
  std::vector<VertexID> multiPoints = mesh.computeMultiPoints();

  MeshPart mPart(mesh, true);
  std::deque<MeshPart> ccs = mPart.getCC();
  for(std::deque<MeshPart>::iterator cc = ccs.begin(); cc != ccs.end(); ++cc) {
    if (((*cc).getGenus() == 0) && ((*cc).getNbBoundaries() == 2)) {
      std::vector<VertexID> mpInsideCC;
      for(std::vector<VertexID>::const_iterator m = multiPoints.begin(); m != multiPoints.end(); ++m)
	if ((*cc).isInsidePoint(*m))
	  mpInsideCC.push_back(*m);

      if (mpInsideCC.size() == 4) {
	VertexID sideA[2];
	VertexID sideB[2];

	sortBySide(mesh, *cc, mpInsideCC, sideA, sideB);

	Mesh ccMesh = (*cc).buildCropMesh();
	MeshPart::IDTranslator idTranslator(*cc);

	VertexID sideALocal[2] = { idTranslator.g2l(sideA[0]), idTranslator.g2l(sideA[1]) };
	VertexID sideBLocal[2] = { idTranslator.g2l(sideB[0]), idTranslator.g2l(sideB[1]) };

	// first parameterization
	CylinderParameterization cParam;
	Mesh ccMeshParam1 = cParam.parameterize3DFixedPoints(ccMesh, sideALocal, sideBLocal);
	FastMarching fMarching;

	PLPath p11 = idTranslator.l2gPLPath(fMarching.getShortestPath(ccMeshParam1, sideALocal[0], sideBLocal[0]), mesh);
	PLPath p12 = idTranslator.l2gPLPath(fMarching.getShortestPath(ccMeshParam1, sideALocal[1], sideBLocal[1]), mesh);

	// second parameterization
	{
	  VertexID t = sideALocal[0];
	  sideALocal[0] = sideALocal[1];
	  sideALocal[1] = t;
	}
	Mesh ccMeshParam2 = cParam.parameterize3DFixedPoints(ccMesh, sideALocal, sideBLocal);

	PLPath p21 = idTranslator.l2gPLPath(fMarching.getShortestPath(ccMeshParam2, sideALocal[0], sideBLocal[0]), mesh);
	PLPath p22 = idTranslator.l2gPLPath(fMarching.getShortestPath(ccMeshParam2, sideALocal[1], sideBLocal[1]), mesh);

	if (p11.length() + p12.length() < p21.length() + p22.length()) {
	  paths.push_back(p11);
	  paths.push_back(p12);
	}
	else {
	  paths.push_back(p21);
	  paths.push_back(p22);
	}
      }
      else if (mpInsideCC.size() == 2) {
	std::cout << "Not yet implemented" << std::endl;
      }
      else if (mpInsideCC.size() == 0)
	std::cout << "Not yet implemented" << std::endl;
      else {
	std::cout << "Number of multipoints: " << mpInsideCC.size() << std::endl;
	throw Exception("buildRectangularCutFromCylindricalTilingFloater(): wrong number of multipoints.");
      }

    }
  }

  return cutMesh(mesh, paths);
}

Mesh MeshPLCut::cutExtremaDiscs(Mesh & mesh, const std::vector<VertexID> & extrema) {
  std::vector<PLPath> paths;
  std::vector<VertexID> multiPoints = mesh.computeMultiPoints();

  MeshPart mPart(mesh, true);
  std::deque<MeshPart> ccs = mPart.getCC();
  for(std::deque<MeshPart>::iterator cc = ccs.begin(); cc != ccs.end(); ++cc) {
    if (((*cc).getGenus() == 0) && ((*cc).getNbBoundaries() == 1)) {
      std::vector<VertexID> mpInsideCC;
      for(std::vector<VertexID>::const_iterator m = multiPoints.begin(); m != multiPoints.end(); ++m)
	if ((*cc).isInsidePoint(*m))
	  mpInsideCC.push_back(*m);

      if (mpInsideCC.size() == 2) {
	std::vector<VertexID> insideExtrema = (*cc).getInsidePoints(extrema);
	if (insideExtrema.size() != 1) {
	  std::cout << "cutExtremaDiscs: wrong number of extrema points" << std::endl;
	  continue;
	}

	Mesh ccMesh = (*cc).buildCropMesh();
	MeshPart::IDTranslator idTranslator(*cc);
	FastMarching fMarching;

	PLPath p1 = idTranslator.l2gPLPath(fMarching.getShortestPath(ccMesh,
								     idTranslator.g2l(mpInsideCC[0]),
								     idTranslator.g2l(insideExtrema.front())), mesh);
	PLPath p2 = idTranslator.l2gPLPath(fMarching.getShortestPath(ccMesh,
								     idTranslator.g2l(insideExtrema.front()),
								     idTranslator.g2l(mpInsideCC[1])), mesh);

	paths.push_back(p2 + p1);
      }
      else {
	std::cout << "cutExtremaDiscs: Not yet implemented" << std::endl;
      }
    }
  }

  return cutMesh(mesh, paths);
}
