/* Taglut : Topological And Geometrical Library - a Useful Toolkit
 * Copyright (C) 2008 Jean-Marie Favreau <jean-marie.favreau@ens-cachan.org>
 *                    CNRS / Univ. Blaise Pascal
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <fstream>
#include <sstream>
#include <assert.h>
#include "FileExceptions.h"
#include "FileManipulator.h"
#include "FileTools.h"

using namespace Taglut;

FileManipulator::FileManipulator() {

}

FileManipulator::FileManipulator(const FileManipulator & fm) : loadFormats(fm.loadFormats), saveFormats(fm.saveFormats),
							       asciiFormat(fm.asciiFormat) {

}

FileManipulator::FileManipulator(const std::string & fileName, const std::string & objectName) {
  load(fileName, objectName);
}

FileManipulator::~FileManipulator() {

}

void FileManipulator::addFormat(const std::string & format, bool ascii) {
  addLoadFormat(format, ascii);
  addSaveFormat(format, ascii);
}

void FileManipulator::addLoadFormat(const std::string & format, bool ascii) {
  loadFormats.push_back(format);
  asciiFormat[format] = ascii;
}

void FileManipulator::addSaveFormat(const std::string & format, bool ascii) {
  saveFormats.push_back(format);
  asciiFormat[format] = ascii;
}


void FileManipulator::load(const std::string & fileName, const std::string & objectName) {
  if (!isLoadFormat(fileName))
    throw Exception("load(2): unknown format");
  loadAltFormat(fileName, FileTools::extension(fileName), objectName);
}

void FileManipulator::load(const char * fileName, const char * objectName) {
  assert(isLoadFormat(fileName));
  if (objectName == NULL)
    load(std::string(fileName));
  else
    load(std::string(fileName), std::string(objectName));
}

bool FileManipulator::isLoadFormat(const std::string & fileName) const {
  std::string ext = FileTools::extension(fileName);

  for(std::vector<std::string>::const_iterator i = loadFormats.begin(); i != loadFormats.end(); ++i)
    if (*i == ext)
      return true;
  return false;
}



void FileManipulator::save(const std::string & fileName, const std::string & objectName) const {
  assert(isSaveFormat(fileName));
  std::string ext = FileTools::extension(fileName);
  saveAltFormat(fileName, ext, objectName);
}

void FileManipulator::save(const char * fileName, const char * objectName) const {
  assert(isSaveFormat(fileName));
  std::string ext = FileTools::extension(fileName);
  if (objectName == NULL)
    saveAltFormat(std::string(fileName), ext);
  else
    saveAltFormat(std::string(fileName), ext, std::string(objectName));
}


bool FileManipulator::isSaveFormat(const std::string & fileName) const {
  std::string ext = FileTools::extension(fileName);

  for(std::vector<std::string>::const_iterator i = saveFormats.begin(); i != saveFormats.end(); ++i)
    if (*i == ext)
      return true;
  return false;
}

bool FileManipulator::isAsciiFormat(const std::string & fileName) const {
  assert(isSaveFormat(fileName) || isLoadFormat(fileName));
  std::string ext = FileTools::extension(fileName);
  std::map<std::string, bool>::const_iterator b = asciiFormat.find(ext);
  if (b == asciiFormat.end())
    return false;
  else
    return (*b).second;
}


std::string FileManipulator::getHeaderVRML() {
  std::ostringstream result;
  result << "#VRML V1.0 ascii" << std::endl << std::endl << "# Generated by FileManipulator" << std::endl << std::endl;
  result << "Separator {" << std::endl << "Switch {" << std::endl;
  return result.str();
}

std::string FileManipulator::getMiddleVRML() {
  std::ostringstream result;

  result << std::endl << "}" << std::endl << "Separator {" << std::endl;

  return result.str();
}

std::string FileManipulator::getFooterVRML() {
  std::ostringstream result;

  result << "}" << std::endl << "}" << std::endl;

  return result.str();
}

std::string FileManipulator::getTransformationsVRML(const std::string & name, unsigned int nbMeshes) {
  std::ostringstream result;

  for(unsigned int i = 0; i < nbMeshes; ++i) {
    result << "\tSeparator {" << std::endl << "\t\tMatrixTransform {" << std::endl << "\t\tmatrix" << std::endl;
    result << "\t\t\t1.000000 0.000000 0.000000 0.000000" << std::endl;
    result << "\t\t\t0.000000 1.000000 0.000000 0.000000" << std::endl;
    result << "\t\t\t0.000000 0.000000 1.000000 0.000000" << std::endl;
    result << "\t\t\t" << "0 0 0 1.000000" << std::endl;
    result << "\t\t}" << std::endl;
    result << "\t\tUSE " << name << "_" << i << std::endl;
    result << "\t}" << std::endl;
  }

  return result.str();
}

bool FileManipulator::checkFormatVRML(std::ifstream & infile) const {
  std::string comment;
  getline(infile, comment);
  return (comment == "#VRML V1.0 ascii") || (comment == "#VRML V1.0 ascii\r");
}

bool FileManipulator::skipBlockVRML(std::ifstream & infile) const {
  std::string buf;
  int nbBrace = 0;

  do {
    if  (infile.eof()) return false;
    infile >> buf;
    if (buf == "{") ++nbBrace;
    if (buf == "}") --nbBrace;
  } while(nbBrace != 0);

  return true;
}

